@@ -1,6 +1,6 @@
 /*-------------------------------------------------------------------------*/
 /*-------------------------------------------------------------------------*
- * isp1161 USB HCD for Linux Version 0.9 (10/4/2001)
+ * isp1161 USB HCD for Linux Version 0.9.5 (10/28/2001)
  * 
  * requires (includes) hc_simple.[hc] simple generic HCD frontend
  *  
@@ -26,6 +26,8 @@
  *
  *-------------------------------------------------------------------------*/
 
+/* V 0.9.5 10/28/2001 correct toggle on error */
+/*         inw, outw macros; ISOC support */ 
 /* V 0.9  remove some bugs, ISOC still missing */
 /* V 0.8.1 9/5/2001 mv memcpy of in packet to hc */
 /* V 0.8 initial release */
@@ -46,6 +48,8 @@
 #include <linux/usb.h>
 
 #undef HC_URB_TIMEOUT
+#undef HC_SWITCH_INT
+#undef HC_ENABLE_ISOC
 
 #include "hc_isp116x.h"
 #include "hc_simple.h"
@@ -59,6 +63,7 @@ static int urb_debug = 0;
 
 static int irq = 10;
 static int hcport = 0x290;
+static int hcport2 = 0x292;
 static int wuport = 0x240;
 
 
@@ -73,18 +78,32 @@ MODULE_PARM(irq,"i");
 MODULE_PARM_DESC(irq,"IRQ 3, 10 (default)");
 MODULE_PARM(hcport,"i");
 MODULE_PARM_DESC(hcport,"ISP116x PORT 0x290");
+MODULE_PARM(hcport2,"i");
+MODULE_PARM_DESC(hcport2,"ISP116x PORT2 0x292");
 MODULE_PARM(wuport,"i");
 MODULE_PARM_DESC(wuport,"WAKEUP PORT 0x240");
 
+static inline void ISP116x_OUTW (int val, int addr) 
+{
+	outw (val, addr);
+	
+}
+
+static inline int ISP116x_INW (int addr)
+{
+	return inw (addr);
+	
+}
+
 static inline int READ_REG32 (hci_t * hci, int regindex)
 {
 	hcipriv_t * hp = &hci->hp; 
 	int val16, val;
 
-	outw (regindex, hp->hcport + 2);
-	val16 = inw (hp->hcport);
+	ISP116x_OUTW (regindex, hp->hcport2);
+	val16 = ISP116x_INW (hp->hcport);
 	val = val16;
-	val16 = inw (hp->hcport);
+	val16 = ISP116x_INW (hp->hcport);
 	val += val16 << 16;
 
 	return val;
@@ -95,8 +114,8 @@ static inline int READ_REG16 (hci_t * hci, int regindex)
 	hcipriv_t * hp = &hci->hp; 
 	int val = 0;
 
-	outw (regindex, hp->hcport + 2);
-	val = inw (hp->hcport);
+	ISP116x_OUTW (regindex, hp->hcport2);
+	val = ISP116x_INW (hp->hcport);
 	
 	return val;
 }
@@ -107,41 +126,47 @@ static inline void READ_REGn16 (hci_t * hci, int regindex, int length, __u8 * bu
 	int i;
 	int val = 0;
 
-	outw (regindex, hp->hcport + 2);
+	ISP116x_OUTW (regindex, hp->hcport2);
 
-	for (i = 0; i < length; i += 2) {
-		val = inw (hp->hcport);
+	for (i = 0; i < length - 1; i += 2) {
+		val = ISP116x_INW (hp->hcport);
 		buffer [i] = val;
 		buffer [i+1] = val >> 8; 
 	}
 	if (length & 1) {
-		val = inw (hp->hcport);
+		val = ISP116x_INW (hp->hcport);
 		buffer [length - 1] = val;
 	}
+#ifdef DEBUG1
+	printk (" READ_REGn16: %d :", length);
+	for (i = 0; i < length; i++)
+		printk (" %2x", buffer [i]);
+	printk("\n");
+#endif
 }
 
 static inline void WRITE_REG32 (hci_t * hci, unsigned int value, int regindex) 
 {
 	hcipriv_t * hp = &hci->hp; 
 
-	outw (regindex | 0x80, hp->hcport + 2);
-	outw (value, hp->hcport);
-	outw (value >> 16, hp->hcport);
+	ISP116x_OUTW (regindex | 0x80, hp->hcport2);
+	ISP116x_OUTW (value, hp->hcport);
+	ISP116x_OUTW (value >> 16, hp->hcport);
 }
 
 static inline void WRITE_REG16 (hci_t * hci, unsigned int value, int regindex) 
 {
 	hcipriv_t * hp = &hci->hp; 
 
-	outw (regindex | 0x80, hp->hcport + 2);
-	outw (value, hp->hcport);
+	ISP116x_OUTW (regindex | 0x80, hp->hcport2);
+	ISP116x_OUTW (value, hp->hcport);
 }
 
 static inline void WRITE_REG0 (hci_t * hci, int regindex) 
 {	
 	hcipriv_t * hp = &hci->hp; 
 
-	outw (regindex | 0x80, hp->hcport + 2);
+	ISP116x_OUTW (regindex | 0x80, hp->hcport2);
 }
 
 static inline void WRITE_REGn16 (hci_t * hci, int regindex, int length, __u8 * buffer) 
@@ -149,14 +174,20 @@ static inline void WRITE_REGn16 (hci_t * hci, int regindex, int length, __u8 * b
 	hcipriv_t * hp = &hci->hp; 
 	int i;
 
-	outw (regindex | 0x80, hp->hcport + 2);
+	ISP116x_OUTW (regindex | 0x80, hp->hcport2);
 
-	for (i = 0; i < length; i+=2) {
-		outw (buffer [i] + (buffer [i+1] << 8), hp->hcport);
+	for (i = 0; i < length - 1; i += 2) {
+		ISP116x_OUTW (buffer [i] + (buffer [i+1] << 8), hp->hcport);
 	}
 	if (length & 1) {
-		outw (buffer [length - 1], hp->hcport);
+		ISP116x_OUTW (buffer [length - 1], hp->hcport);
 	}
+#ifdef DEBUG1
+	printk (" WRITE_REGn16: %d :", length);
+	for (i = 0; i < length; i++)
+		printk (" %2x", buffer [i]);
+	printk("\n");
+#endif
 }
 
 /*-------------------------------------------------------------------------*/
@@ -174,20 +205,29 @@ static inline void hc_mark_last_trans (hci_t * hci)
 static inline void hc_flush_data_cache (hci_t * hci, void * data, int len) 
 {}
 
-#define PARTS 4
-
 static inline int hc_add_trans (hci_t * hci, int len, void * data,
 		int toggle, int maxps, int slow, int endpoint, int address, int pid, int format)
 {
 	hcipriv_t * hp = &hci->hp; 
 	int last = 0;
 	__u8 * ptd = hp->tl;
+	int parts = 2;
 
 	/* just send 4 packets of each kind at a frame,
 	 * other URBs also want some bandwitdh, and a NACK is cheaper
 	 * with less packets */
-	if (len > maxps * PARTS)
-			len = maxps * PARTS; 
+	
+	switch (hci->active_urbs) {
+		case 1: parts =  8; break;
+		case 2: parts =  6; break;
+		case 3: parts =  4; break;
+		case 4: parts =  3; break;
+		case 5: parts =  2; break;
+		default: parts = 2;
+	}
+	
+	if (len > maxps * parts)
+			len = maxps * parts; 
 	
 	if (hp->units_left < ((len + 8 + 3) & ~0x3))
 		return -1;
@@ -235,6 +275,9 @@ static inline int hc_parse_trans (hci_t * hci, int * actbytes, void * data,
 		memcpy (data, ptd + 8, *actbytes);
 	
 	*toggle = (ptd [1] >> 2 & 1);
+	if (*cc > 0 && *cc < 0xE && *cc != 9)
+		*toggle = !*toggle;
+
 /*	*active = (ptd[1] >> 3) & 0x1;	
 	*endpoint = ptd [3] >> 4;
 	*address = ptd [6] & 0x7f;
@@ -248,31 +291,21 @@ static inline int hc_parse_trans (hci_t * hci, int * actbytes, void * data,
 
 /*-------------------------------------------------------------------------*/
 
-/*
-static void hc_start_list (hci_t * hci, ed_t * ed)
+static void hc_start_int (hci_t * hci)
 {
-	hcipriv_t * hp = hci->hcipriv;
-	int units_left = hp->atl_buffer_len;
-	
-// printk("hc_start_list atl_buffer_len: %d atl_len: %d\n", hp->atl_buffer_len, hp->atl_len);
-	switch (usb_pipetype (ed->pipe)) {
-	case PIPE_CONTROL:
-	case PIPE_BULK:
-	case PIPE_INTERRUPT:
-		if (hp->atl_len == 0) {
-			hp->tlp = 0;
-			sh_schedule_trans (hci, TR_INTR | TR_CTRL | TR_BULK, &units_left);
-			mark_last_trans (hci);
-			hp->atl_len = hp->tlp;
-			if (hp->atl_len > 0) {
-// printk("hc_start_list: send %d bytes\n", hp->atl_len);
-				WRITE_REG16 (hci, hp->atl_len, HcTransferCounter);
-				WRITE_REGn16 (hci, HcATLBufferPort, hp->atl_len, hp->tl);
-			}
-		}
-	}
+#ifdef HC_SWITCH_INT	
+	int mask = SOFITLInt | ATLInt | OPR_Reg;
+	WRITE_REG16 (hci, mask, HcuPInterrupt); 
+	WRITE_REG16 (hci, mask, HcuPInterruptEnable);
+#endif
 }	
-*/
+
+static void hc_stop_int (hci_t * hci)
+{
+#ifdef HC_SWITCH_INT
+	WRITE_REG16 (hci, 0, HcuPInterruptEnable);
+#endif
+}
 
 /* an interrupt happens */
 static void hc_interrupt (int irq, void * __hci, struct pt_regs * r)
@@ -280,6 +313,7 @@ static void hc_interrupt (int irq, void * __hci, struct pt_regs * r)
 	hci_t * hci = __hci;
 	hcipriv_t * hp = &hci->hp; 
 	int ints_uP, ints = 0, bstat = 0;
+	int iso_buffer_index = 0;
 	
 	if ((ints_uP = (READ_REG16 (hci, HcuPInterrupt) & READ_REG16 (hci, HcuPInterruptEnable))) == 0) {
 // printk("int1 %x %x bstat %x\n", ints_uP, ints, bstat);
@@ -300,18 +334,49 @@ static void hc_interrupt (int irq, void * __hci, struct pt_regs * r)
 		WRITE_REG32 (hci, ints, HcInterruptStatus);
 		WRITE_REG32 (hci, OHCI_INTR_MIE, HcInterruptEnable);
 	}
+	hci->frame_number = GET_FRAME_NUMBER (hci);
 
 	bstat = READ_REG16 (hci, HcBufferStatus);
-	if (ints_uP & SOFITLInt) {	
-/*		READ_REGn16 (hci,reg, len, hci->tl);
+	if (ints_uP & SOFITLInt) {			
+		hci->frame_number = GET_FRAME_NUMBER (hci);
+
+#ifdef HC_ENABLE_ISOC
+	/* read itl1 first. if there is an buffer overflow HC will stop at itl1 
+	   There will be a quirks in sh_done_list if this happens !!!!! */
+
+		if ((bstat & ITL1BufferFull) && (bstat & ITL1BufferDone)) {
+			hp->itl1_len = READ_REG16 (hci, HcReadBackITL1Length);
+			if (hp->itl1_len > 0) {
+				WRITE_REG16 (hci, hp->itl1_len, HcTransferCounter);
+				READ_REGn16 (hci, HcITLBufferPort, hp->itl1_len, hp->tl);
+				hp->tlp = 0;
+				hci->td_array = &hci->i_td_array [1];
+				sh_done_list (hci);
+			}
+			iso_buffer_index = 1;
+		}
+		if ((bstat & ITL0BufferFull) && (bstat & ITL0BufferDone)) {
+			hp->itl0_len = READ_REG16 (hci, HcReadBackITL0Length);
+			if (hp->itl0_len > 0) {
+				WRITE_REG16 (hci, hp->itl0_len, HcTransferCounter);
+				READ_REGn16 (hci, HcITLBufferPort, hp->itl0_len, hp->tl);
+				hp->tlp = 0;
+				hci->td_array = &hci->i_td_array [0];
+				sh_done_list (hci);
+			}
+			iso_buffer_index = 0;
+		}
 		hp->tlp = 0;
-		done_list (hci);
-		schedule_trans (hci, TR_ISO);
-		WRITE_REG16 (hci, hp->tlp, HcTransferCounter);
-		WRITE_REGn16 (hci, HcITLBufferPort, hp->tlp, hp->tl);
-*/	}
+		hci->td_array = &hci->i_td_array [iso_buffer_index];
+		sh_scan_iso_urb_list (hci, &hci->iso_list, hci->frame_number + 1);
+		if (hp->tlp >0) {
+			WRITE_REG16 (hci, hp->tlp, HcTransferCounter);
+			WRITE_REGn16 (hci, HcITLBufferPort, hp->tlp, hp->tl);
+		}
+#endif
+	}
+
 
-	hci->frame_number = GET_FRAME_NUMBER (hci);
 //	if (hci->frame_no == 0) 
 //		printk("*** INTERRUPT frame %u intuP %x ints %x bstat %x \n", hci->frame_no, ints_uP, ints, bstat);
 
@@ -319,19 +384,20 @@ static void hc_interrupt (int irq, void * __hci, struct pt_regs * r)
 // printk("int2 %x %x bstat %x\n", ints_uP, ints, bstat);
 		if ((bstat & ATLBufferFull) && (bstat & ATLBufferDone)) {
 			if (hp->atl_len > 0) {
+				WRITE_REG16 (hci, hp->atl_len, HcTransferCounter);
 				READ_REGn16 (hci, HcATLBufferPort, hp->atl_len, hp->tl);
 				hp->tlp = 0;
+				hci->td_array = &hci->a_td_array;
 				sh_done_list (hci);
 			}
 			hp->tlp = 0;
-			hci->trans = 0;
 		}
 	}
-	if (!hci->trans)
+	if (!hci->td_array->len)
 		sh_del_list (hci);
-
+	hci->td_array = &hci->a_td_array;
 	bstat = READ_REG16 (hci, HcBufferStatus);
-	if (hci->trans == 0 && !(bstat & ATLBufferFull)) {
+	if (hci->td_array->len == 0 && !(bstat & ATLBufferFull)) {
 		hp->units_left = hp->atl_buffer_len;
 		hp->tlp = 0;
 		sh_schedule_trans (hci);
@@ -446,6 +512,9 @@ static int hc_start (hci_t * hci)
 	WRITE_REG32 (hci, mask, HcInterruptStatus);
 
 	mask = SOFITLInt | ATLInt | OPR_Reg;
+#ifdef HC_SWITCH_INT
+	mask = 0;
+#endif
 	WRITE_REG16 (hci, mask, HcuPInterrupt); 
 	WRITE_REG16 (hci, mask, HcuPInterruptEnable);
 
@@ -486,6 +555,12 @@ static hci_t * __devinit hc_alloc_hci (void)
 	
 	hp->irq = -1;
 	hp->hcport = -1;
+
+ 	hci->a_td_array.len = 0;
+ 	hci->i_td_array[0].len = 0;
+ 	hci->i_td_array[1].len = 0;
+ 	hci->td_array = &hci->a_td_array;
+ 	hci->active_urbs = 0;
  
 	INIT_LIST_HEAD (&hci->hci_hcd_list);
 	list_add (&hci->hci_hcd_list, &hci_hcd_list);
@@ -538,10 +613,15 @@ static void hc_release_hci (hci_t * hci)
 		kfree (hp->tl);
 
 	if (hp->hcport > 0) {
-		release_region (hp->hcport, 4);
+		release_region (hp->hcport, 2);
 		hp->hcport = 0;
 	}
 
+	if (hp->hcport2 > 0) {
+		release_region (hp->hcport2, 2);
+		hp->hcport2 = 0;
+	}
+
 	if (hp->wuport > 0) {
 		release_region (hp->wuport, 2);
 		hp->wuport = 0;
@@ -567,7 +647,7 @@ static void hc_release_hci (hci_t * hci)
  * return success. */
 
  
-static int __devinit hc_found_hci (int addr, int wuaddr, int irq, int dma)
+static int __devinit hc_found_hci (int addr, int addr2, int wuaddr, int irq, int dma)
 {
 	hci_t * hci;
 	hcipriv_t * hp; 
@@ -579,7 +659,7 @@ static int __devinit hc_found_hci (int addr, int wuaddr, int irq, int dma)
 
 	hp = &hci->hp;
 	
-	if (!request_region (addr, 4, "ISP116x USB HOST")) {
+	if (!request_region (addr, 2, "ISP116x USB HOST")) {
 		err ("request address %d-%d failed", addr, addr+4);
 		hc_release_hci (hci);
 		return -EBUSY;	
@@ -587,6 +667,14 @@ static int __devinit hc_found_hci (int addr, int wuaddr, int irq, int dma)
 
 	hp->hcport = addr;
 
+	if (!request_region (addr2, 2, "ISP116x USB HOST")) {
+		err ("request address %d-%d failed", addr, addr+4);
+		hc_release_hci (hci);
+		return -EBUSY;	
+	}
+
+	hp->hcport2 = addr2;
+
 	if ((READ_REG16(hci, HcChipID) & 0xff00) != 0x6100) {
 		hc_release_hci (hci);
 		return -ENODEV;
@@ -610,8 +698,8 @@ static int __devinit hc_found_hci (int addr, int wuaddr, int irq, int dma)
 		return -ENOMEM;
 	}
 
-	printk(KERN_INFO __FILE__ ": USB ISP116x at %x/%x IRQ %d Rev. %x ChipID: %x\n",
-		addr, wuaddr, irq, READ_REG32(hci, HcRevision), READ_REG16(hci, HcChipID));
+	printk(KERN_INFO __FILE__ ": USB ISP116x at %x/%x,%x IRQ %d Rev. %x ChipID: %x\n",
+		addr, addr2, wuaddr, irq, READ_REG32(hci, HcRevision), READ_REG16(hci, HcChipID));
 
 	/* FIXME this is a second HC reset; why?? */
 	WRITE_REG32 (hci, hp->hc_control = OHCI_USB_RESET, HcControl);
@@ -626,23 +714,13 @@ static int __devinit hc_found_hci (int addr, int wuaddr, int irq, int dma)
 		return -EBUSY;
 	}
 	hp->irq = irq;
-#if 0
-	if (dma && !request_dma (dma, "ISP116x USB HOST")) {
-		err ("request dma %d failed", dma);
-		dma = 0;	
-	}
 
-	hp->dmahci->dmaci->dma = dma;
-#endif
 	if (hc_start (hci) < 0) {
 		err ("can't start usb-%x", addr);
 		hc_release_hci (hci);
 		return -EBUSY;
 	}
 
-#ifdef	DEBUG
-//	hci_dump (hci, 1);
-#endif
 	return 0;
 }
 
@@ -653,7 +731,7 @@ static int __init hci_hcd_init (void)
 {
 	int ret;
 
-	ret = hc_found_hci (hcport, wuport, irq, 0);   
+	ret = hc_found_hci (hcport, hcport2, wuport, irq, 0);   
 
 	return ret;
 }
