@@ -1,6 +1,6 @@
 /*-------------------------------------------------------------------------*/
 /*-------------------------------------------------------------------------*
- * simple generic USB HCD frontend Version 0.9 (9/23/2001)
+ * simple generic USB HCD frontend Version 0.9.5 (10/28/2001)
  * for embedded HCs (eg. isp1161, MPC850, ...)
  * 
  * USB URB handling, hci_ hcs_
@@ -25,6 +25,14 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  *
  *-------------------------------------------------------------------------*/
+/* 
+ * V0.9.5 (10/28/2001) start_int()/stop_int() enable interrupt processing on demand
+ *        ISOC processing
+ * V0.9.2 (10/10/2001) sh_done_list: on error just reject last packet
+ * V0.9.1 (10/8/2001)  unlink_urb check also for USB_ST_URB_PENDING
+ * V0.9 (9/23/2001) 
+ * * * * * * * */
+
 
 /* main lock for urb access */
 static spinlock_t usb_urb_lock = SPIN_LOCK_UNLOCKED; 
@@ -165,7 +173,7 @@ static int hci_unlink_urb (urb_t * urb)
 		
 	spin_lock_irqsave (&usb_urb_lock, flags);
 
-	if (!list_empty (&urb->urb_list)) { /* URB active? */
+	if (!list_empty (&urb->urb_list) && urb->status == USB_ST_URB_PENDING) { /* URB active? */
 	
 		if (urb->transfer_flags & (USB_ASYNC_UNLINK | USB_TIMEOUT_KILLED)) { 
 			/* asynchron with callback */
@@ -175,21 +183,23 @@ static int hci_unlink_urb (urb_t * urb)
 			spin_unlock_irqrestore (&usb_urb_lock, flags);
 
 		} else { /* synchron without callback */
-			
+
 			add_wait_queue (&hci->waitq, &wait);	  
 
 			set_current_state (TASK_UNINTERRUPTIBLE);
 			comp = urb->complete;
 			urb->complete = NULL;
-
+	
 			list_del (&urb->urb_list); /* relink the urb to the del list */
 			list_add (&urb->urb_list, &hci->del_list);
+
 			spin_unlock_irqrestore (&usb_urb_lock, flags);
 			
 			schedule_timeout(HZ/50);
-			
+
 			if (!list_empty (&urb->urb_list))
 				list_del (&urb->urb_list);
+	
 			urb->complete = comp;
 			urb->hcpriv = NULL;
 			remove_wait_queue (&hci->waitq, &wait); 
@@ -360,7 +370,6 @@ static inline void qu_queue_active_urb (hci_t * hci, urb_t * urb, epd_t * ed)
 		
 		case PIPE_ISOCHRONOUS:
 			list_add (&urb->urb_list, &hci->iso_list);
-			urb_state = urb->number_of_packets;
 			break;
 	}
 
@@ -392,7 +401,10 @@ static int qu_queue_urb (hci_t * hci, urb_t * urb)
 	 } else {
 	 	ed->pipe_head = urb;
 		qu_queue_active_urb (hci, urb, ed);
+		if (++hci->active_urbs == 1)
+			hc_start_int (hci);
 	}
+
 	return 0;
 }
 
@@ -409,20 +421,26 @@ static urb_t * qu_next_urb (hci_t * hci, urb_t * urb, int resub_ok)
 	list_del (&urb->urb_list);
 	INIT_LIST_HEAD (&urb->urb_list);
 
+	if (ed->pipe_head == urb) {
+
 #ifdef HC_URB_TIMEOUT
-	if (urb->timeout)
-		del_timer (&ed->timeout);
+		if (urb->timeout)
+			del_timer (&ed->timeout);
 #endif
-	
-	if (!list_empty (&ed->urb_queue)) {
-		urb = list_entry (ed->urb_queue.next, urb_t, urb_list);
-		list_del (&urb->urb_list);
-		INIT_LIST_HEAD (&urb->urb_list);
-		ed->pipe_head = urb;
-		qu_queue_active_urb (hci, urb, ed);
-	} else {
-		ed->pipe_head = NULL;
-		urb = NULL;
+
+		if (!--hci->active_urbs)
+			hc_stop_int (hci);
+
+		if (!list_empty (&ed->urb_queue)) {
+			urb = list_entry (ed->urb_queue.next, urb_t, urb_list);
+			list_del (&urb->urb_list);
+			INIT_LIST_HEAD (&urb->urb_list);
+			ed->pipe_head = urb;
+			qu_queue_active_urb (hci, urb, ed);
+		} else {
+			ed->pipe_head = NULL;
+			urb = NULL;
+		}
 	}
 	return urb;
 }
@@ -440,20 +458,21 @@ static urb_t * qu_return_urb (hci_t * hci, urb_t * urb, int resub_ok)
 /* SCHEDULE operations
  * 
  * * * */
-static int sh_scan_iso_urb_list (hci_t * hci, struct list_head * list_lh) 
+static int sh_scan_iso_urb_list (hci_t * hci, struct list_head * list_lh, int frame_number) 
 {
 	struct list_head * lh = list_lh->next;
 	urb_t * urb;
-	
+
+	hci->td_array->len = 0;
 
 	while  (lh != list_lh) {
 		urb = list_entry (lh, urb_t, urb_list);
 		lh = lh->next;
-		if (((hci->frame_number - urb->start_frame) & 0x7ff) < urb->number_of_packets) {
+		if (((frame_number - urb->start_frame) & 0x7ff) < urb->number_of_packets) {
 			if (!sh_add_packet (hci, urb))
 				return 0;
 			else {
-				if (((hci->frame_number - urb->start_frame) & 0x7ff) > 0x400) {
+				if (((frame_number - urb->start_frame) & 0x7ff) > 0x400) {
 					if (qu_urbstate(urb) > 0)
 						urb = qu_return_urb (hci, urb, 1);
 					else
@@ -476,7 +495,7 @@ static int sh_scan_urb_list (hci_t * hci, struct list_head * list_lh)
 		urb = list_entry (lh, urb_t, urb_list);
 		if (!usb_pipeint (urb->pipe) ||
 			(((hci->frame_number - urb->start_frame) & 0x7ff) >= urb->interval)) {
-				if (!sh_add_packet (hci, urb))
+					if (!sh_add_packet (hci, urb))
 					return 0;
 				else {
 //					printk("INT: start: %d fr_no: %d int: %d pint: %d\n", urb->start_frame, hci->frame_number, urb->interval, usb_pipeint (urb->pipe));
@@ -496,7 +515,7 @@ static int sh_schedule_trans (hci_t * hci)
 	int units_left;
 	struct list_head * lh;
 
-	hci->trans = 0;
+	hci->td_array->len = 0;
 
 	units_left = sh_scan_urb_list (hci, &hci->intr_list);
 
@@ -597,10 +616,10 @@ static int sh_add_packet (hci_t * hci, urb_t * urb)
   	
 #endif
 	if (ret >= 0) {
-		hci->trans_array [hci->trans].urb = urb;
-		hci->trans_array [hci->trans].len = ret;
-		hci->trans_array [hci->trans].iso_index = i;
-		hci->trans ++;
+		hci->td_array->td [hci->td_array->len].urb = urb;
+		hci->td_array->td [hci->td_array->len].len = ret;
+		hci->td_array->td [hci->td_array->len].iso_index = i;
+		hci->td_array->len ++;
 		return 1;
 	}
 	return 0;
@@ -627,23 +646,22 @@ static int sh_done_list (hci_t * hci)
 	int len;
 	int iso_index = 0;
 
-	for (trans = 0; ret && trans < hci->trans && trans < MAX_TRANS; trans++) {
+	for (trans = 0; ret && trans < hci->td_array->len && trans < MAX_TRANS; trans++) {
 
-		urb = hci->trans_array [trans].urb;
-		len = hci->trans_array [trans].len;
+		urb = hci->td_array->td [trans].urb;
+		len = hci->td_array->td [trans].len;
 		
 		if (usb_pipeisoc (urb->pipe)) {
-			iso_index = hci->trans_array [trans].iso_index;
-			urb_state = urb->number_of_packets - iso_index;
+			iso_index = hci->td_array->td [trans].iso_index;
 			data = urb->transfer_buffer + 
 					urb->iso_frame_desc [iso_index].offset;
 			toggle = 0;
-		} else {
-			urb_state = qu_urbstate (urb);
+		} else {	
 			data = urb->transfer_buffer + urb->actual_length;
 			toggle = usb_gettoggle (urb->dev, usb_pipeendpoint (urb->pipe), 
 					usb_pipeout (urb->pipe));
 		}
+		urb_state = qu_urbstate (urb);
 		ret = hc_parse_trans (hci, &actbytes, data, &cc, &toggle, len);
 		maxps = usb_maxpacket (urb->dev, urb->pipe, usb_pipeout (urb->pipe));
 
@@ -667,32 +685,32 @@ printk("hc USBx active %d, cc %x, urb_state %x, ret %x\n", active, cc, urb_state
 								cc = 0;
 			if (cc) { /* last packet has an error */
 				if (hc_error_verbose)
-					printk("hc USB %d error %d in frame %d ep %d out %d addr %d\n", 
+					printk("hc USB %d error %d in frame %d ep %d out %d addr %d au: %d\n", 
 						urb->error_count, cc, hci->frame_number, 
 							usb_pipeendpoint (urb->pipe),
-							usb_pipeout (urb->pipe), usb_pipedevice (urb->pipe));
+							usb_pipeout (urb->pipe), usb_pipedevice (urb->pipe), hci->active_urbs);
 
 				if (++urb->error_count > 3 || cc == TD_CC_STALL) {
 					active = 0;
 					urb_state = 0; /* return */
 				} else { 
-			//		actbytes = actbytes & ~(maxps - 1);
-					actbytes = 0;
 					active = 1;
 				}
-			} else if (urb_state != US_CTRL_SETUP) { /* no error */
-//				if (!usb_pipeout (urb->pipe) && actbytes && data_urb != data)
-//					memcpy (data_urb, data, actbytes);
+//				actbytes = actbytes & ~(maxps - 1);
+
+			} else
+				urb->error_count = 0;
+
+			if (urb_state != US_CTRL_SETUP) { /* no error */
 				urb->actual_length += actbytes;
 				usb_settoggle (urb->dev, usb_pipeendpoint (urb->pipe), 
 					usb_pipeout (urb->pipe), 
 					toggle);
-//					(!actbytes || !((actbytes - 1) & maxps) ? !toggle : toggle));
 			}
 			if (usb_pipeisoc (urb->pipe)) {
 				urb->iso_frame_desc [iso_index].actual_length = actbytes;
 				urb->iso_frame_desc [iso_index].status = cc_to_error [cc];
-				active = 0;
+				active = (iso_index < urb->number_of_packets);
 			}
 			if (!active) {
 				if (!(urb_state--)) {
@@ -702,9 +720,12 @@ printk("hc USBx active %d, cc %x, urb_state %x, ret %x\n", active, cc, urb_state
 					qu_seturbstate (urb, urb_state);
 				}
 			}
-		}
+		} else
+			urb->error_count = 0;
 	}	
 
+	hci->td_array->len = 0;
+	
 	return ret;
 }
 
